/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

package net.jasonfreeman.itsm;

import java.awt.Container;
import java.io.File;

import javax.swing.JApplet;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import quicktime.QTException;
import quicktime.std.movies.Movie;
import net.jasonfreeman.itsm.analysis.Analyzer;
import net.jasonfreeman.itsm.analysis.Chunk;
import net.jasonfreeman.itsm.gui.CanceledPanel;
import net.jasonfreeman.itsm.gui.FinishedPanel;
import net.jasonfreeman.itsm.gui.LicensePanel;
import net.jasonfreeman.itsm.gui.LoadingPanel;
import net.jasonfreeman.itsm.gui.ProgressManager;
import net.jasonfreeman.itsm.gui.SharePanel;
import net.jasonfreeman.itsm.gui.SongSelectionPanel;
import net.jasonfreeman.itsm.gui.StructurePanel;
import net.jasonfreeman.itsm.gui.iTunesXMLPanel;
import net.jasonfreeman.itsm.mixing.Exporter;
import net.jasonfreeman.itsm.mixing.Mixer;
import net.jasonfreeman.itsm.selection.FileSelector;
import net.jasonfreeman.itsm.selection.Song;
import net.jasonfreeman.itsm.util.ErrorHandler;
import net.jasonfreeman.itsm.util.QTJavaCheck;
import net.jasonfreeman.itsm.util.QTSessionManager;
import net.jasonfreeman.itsm.util.RemoteDatabase;
import net.jasonfreeman.itsm.util.SettingsManager;

/**
 * The main ITSM applet.
 * 
 * @author jason
 */
public class ITSM extends JApplet {

	private Chunk[] chunks;

	private boolean contentPaneLocked = false;

	private Container waitingContentPane = null;

	private RemoteDatabase db = null;
	
	private boolean executing = false;
	
	private boolean firstRun = true;
	
	private FinishedPanel finishedPanel = null;

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.applet.Applet#init()
	 */
	public void init() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.applet.Applet#start()
	 */
	public void start() {
		try {
			// start things up
			ErrorHandler.getInstance().setApplet(this);
			// get something up on the GUI asap...
			final LoadingPanel loading = new LoadingPanel(this, "Please wait while iTSM loads...");
			// do everything else in a separate thread so GUI will display
			final ITSM applet = this;
			Runnable t = new Runnable() {
				public void run() {
					try {
						// some browsers don't properly stop applet when page is
						// refreshed
						loading.setProgress(5);
						shutdownSingletons(true);
						loading.setProgress(10);
						db = new RemoteDatabase();
						loading.setProgress(60);
						if (!QTJavaCheck.checkQTJVersion()) {
							return;
						}
						loading.setProgress(75);
						// get settings from GUI
						showWizard();
						getRemoteDatabase().sendParameter("session_state",
								"wizard");
					} catch (Throwable e) {
						ErrorHandler.getInstance().handleError(e);
					}
				}
			};
			new Thread(t).start();
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
		}
	}

	/**
	 * Once GUI choices are made, create the actual signature.
	 *
	 */
	public void finish() {
		try {
			executing = true;
			SettingsManager settings = SettingsManager.getInstance();
			ProgressManager progress = ProgressManager.getInstance();
			settings.setProcessing(true);
			getRemoteDatabase().sendParameter("session_state", "processing");
			final ITSM applet = this;
			Runnable updateAComponent = new Runnable() {
				public void run() {
					ProgressManager.getInstance().buildProgressDialog(applet);
				}
			};
			SwingUtilities.invokeLater(updateAComponent);

			// get the files we want to work with
			progress.setTaskInfo("Analyzing iTunes database");
			progress.setPercentInfo(0);
			progress.setTotalInfo(0);
			if (!settings.isProcessing()) {
				executing = false;
				return;
			}
			int maxSongs = settings.getMaxSongs();
			Song[] songs = new Song[maxSongs];
			FileSelector fs = new FileSelector(settings.getiTunesXmlFile(),
					maxSongs);
			songs = fs.selectFiles();
			if (songs == null) {
				return;
			}

			// do the analysis
			progress.setTaskInfo("Analyzing songs");
			progress.setPercentInfo(0);
			progress.setTotalInfo(2);
			if (!settings.isProcessing()) {
				executing = false;
				return;
			}
			Analyzer analyzer = new Analyzer();
			chunks = analyzer.doAnalysis(songs);
			if ((chunks == null) || !settings.isProcessing()) {
				executing = false;
				return;
			}
			// do the mixing
			progress.setTaskInfo("Mixing signature");
			progress.setTotalInfo(98);
			if (!settings.isProcessing()) {
				executing = false;
				return;
			}
			Mixer mixer = new Mixer();
			Movie result = mixer.mixChunks(chunks);
			if (result == null) {
				executing = false;
				QTSessionManager.shutdown();
				return;
			}

			// export to file
			progress.setTaskInfo("Saving signature");
			progress.setTotalInfo(99);
			progress.setPercentInfo(0);
			if (!settings.isProcessing()) {
				executing = false;
				QTSessionManager.shutdown();
				return;
			}
			try {
				if (!Exporter.export(result)) {
					mixer.cleanup();
					SettingsManager.getInstance().setProcessing(false);
					executing = false;
					new CanceledPanel(applet);
					return;
				} else {
					mixer.cleanup();
				}
			} catch (QTException e) {
				ErrorHandler.getInstance().handleError(e);
				executing = false;
				QTSessionManager.shutdown();
				return;
			}

			// finish up
			progress.setTaskInfo("Finished");
			progress.setTotalInfo(100);
			progress.setPercentInfo(100);
			settings.setProcessing(false);
			getRemoteDatabase().sendParameter("session_state", "done");
			finishedOptions();
			executing = false;
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
			executing = false;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.applet.Applet#stop()
	 */
	public void stop() {
		// if movie is playing when QT shuts down, we'll get exceptions and sometimes a crash!
		if (finishedPanel != null) {
			finishedPanel.stopMovie();
		}
		try {
			shutdownSingletons(false);
		} catch (Throwable e) {
			ErrorHandler.getInstance().handleError(e);
		}
	}

	public void restart() {
		new LoadingPanel(this, "Please wait while iTSM loads...");
		while (executing) {
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
		}
		SettingsManager settings = SettingsManager.getInstance();
		settings.updateOutFileName();
		ProgressManager.shutdown();
		firstRun = false;
		start();
	}

	/**
	 * Show GUI options once signature is created.
	 *
	 */
	public void finishedOptions() {
		final ITSM applet = this;
		Runnable updateAComponent = new Runnable() {
			public void run() {
				try {
					new FinishedPanel(applet);
				} catch (Throwable e) {
					ErrorHandler.getInstance().handleError(e);
				}
			}
		};
		SwingUtilities.invokeLater(updateAComponent);
	}

	/**
	 * Show GUI for user to upload signature to gallery.
	 *
	 */
	public void share() {
		final ITSM applet = this;
		Runnable updateAComponent = new Runnable() {
			public void run() {
				try {
					new SharePanel(applet);
				} catch (Throwable e) {
					ErrorHandler.getInstance().handleError(e);
				}
			}
		};
		SwingUtilities.invokeLater(updateAComponent);
	}

	/**
	 * Show GUI with structure of signature contents.
	 *
	 */
	public void showStructure() {
		final ITSM applet = this;
		Runnable updateAComponent = new Runnable() {
			public void run() {
				try {
					new StructurePanel(applet, chunks);
				} catch (Throwable e) {
					ErrorHandler.getInstance().handleError(e);
				}
			}
		};
		SwingUtilities.invokeLater(updateAComponent);
	}

	public void shutdownSingletons(boolean wait) {
		SettingsManager settings = SettingsManager.getInstance();
		settings.setProcessing(false);
		while (wait && executing) {
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {
			}
		}
		SettingsManager.shutdown();
		ErrorHandler.shutdown();
		ProgressManager.shutdown();
		if (db != null)
			db.shutdown(wait);
	}

	public void setContentPane(Container contentPane) {
		if (!contentPaneLocked)
			super.setContentPane(contentPane);
		else
			waitingContentPane = contentPane;
	}

	/**
	 * When the content pane is locked, calls to setContentPane are ignored
	 * until the content pane is unlocked. At that point, whatever the last contentPane
	 * request to come in is handled.
	 * 
	 * @param locked
	 */
	public void setContentPaneLocked(boolean locked) {
		contentPaneLocked = locked;
		if (!locked && (waitingContentPane != null)) {
			setContentPane(waitingContentPane);
			validate();
		}
	}

	public RemoteDatabase getRemoteDatabase() {
		return db;
	}
	
	public boolean isFirstRun() {
		return firstRun;
	}
	
	public void setFinishedPanel(FinishedPanel finishedPanel) {
		this.finishedPanel = finishedPanel;
	}
	
	public void showWizard() {
        if (isFirstRun()) {
        		// only show license agreement the first time
    			new LicensePanel("/html/license.html", this);
        } else {
            // only add XML pane if we can't locate the file automatically
            String xmlFile = SettingsManager.getInstance().getiTunesXmlFile();
            File f = new File(xmlFile);
            if (!f.exists()) {
            		new iTunesXMLPanel(this);
            } else {
                 new SongSelectionPanel(this);
            }
        }
	}
}