/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.analysis;

import java.net.MalformedURLException;


import net.jasonfreeman.itsm.selection.Song;
import net.jasonfreeman.itsm.util.ErrorHandler;
import net.jasonfreeman.itsm.util.QTSessionManager;
import net.jasonfreeman.itsm.util.SettingsManager;
import quicktime.QTException;
import quicktime.QTNullPointerException;
import quicktime.sound.SoundConstants;
import quicktime.std.StdQTConstants;
import quicktime.std.StdQTException;
import quicktime.std.movies.Movie;
import quicktime.std.movies.media.DataRef;
import quicktime.std.movies.media.SoundDescription;
import quicktime.std.qtcomponents.MovieExporter;
import quicktime.util.QTHandle;
import quicktime.util.QTUtils;

/**
 * @author jason
 * 
 * Decode an audio file to 8 bit, 8 kHz, mono AIFF.
 * 
 */
public class Decoder {

	Movie orig;

	SettingsManager settings = SettingsManager.getInstance();

	int position = 0;

	int duration, offset;

	MovieExporter theMovieExp;

	int segmentSize = 6000;
	
	int byteSize;
	
	int byteClumps;
	
	byte[] bytes;
	
	int byteClumpIndex;
	
	Song song;
	
	int segmentNumber = -1;
	
	QTHandle handle = null;
	
	double hopRatio;
	
	DataRef ref = null;
	
	int[] markers;
	
	int origTimeBase;
	
	private SoundDescription sd;

	public Decoder(Song song, float rank) throws QTException, MalformedURLException {
		this.song = song;
		hopRatio = SettingsManager.getInstance().getHopRatio();
		byteSize = SettingsManager.getInstance().getFftFrameGroupSize(rank) * SettingsManager.getInstance().getFftFrameSize();
		byteClumps = 25000 / SettingsManager.getInstance().getFftFrameGroupSize(rank);
		bytes = new byte[byteSize];
		byteClumpIndex = byteClumps;
		orig = song.openAsMovie();
		origTimeBase = orig.getTimeScale();
		duration = orig.getDuration();
		markers = new int[3]; // in fastMode, we don't analyze the whole song, just segments starting from a few markers
		for (int i=0; i < markers.length; i++) {
			markers[i] = duration * i / markers.length;
		}
	}

	private void allocateExporter() throws QTException, MalformedURLException {
		orig = song.openAsMovie();
		origTimeBase = orig.getTimeScale();
		duration = orig.getDuration();
		theMovieExp = new MovieExporter(QTUtils.toOSType("snd "));
		sd = new SoundDescription(0);
		sd.setNumberOfChannels(1);
		sd.setSampleRate(8000);
		sd.setSampleSize(8);
		sd.setDataFormat(SoundConstants.k8BitOffsetBinaryFormat);
		sd.lock();
		theMovieExp.setSampleDescription(sd, StdQTConstants.soundMediaType);

	}
	
	public byte[] getNextSegment() {
		try {
			segmentNumber++;
			if ((position + segmentSize) > duration)
				return null;
			if (byteClumpIndex < byteClumps) {
				int bytesLeft = handle.getSize() - offset - byteClumpIndex * byteSize;
				if (bytesLeft < byteSize)
					return null;
				try {
					handle.copyToArray(offset + byteClumpIndex * byteSize, bytes, 0, byteSize);
					convertByteFormat(bytes);
				} catch (Exception e) {
					ErrorHandler.getInstance().handleError(e);
					return null;
				}
				byteClumpIndex++;
				position += (segmentSize * hopRatio);
				return bytes;
			}
			if (settings.isFastMode()) {
				int i=0;
				while ((i < markers.length) && (position > markers[i]))
					i++;
				if (i == markers.length)
					return null;
				position = markers[i];
			}
			byteClumpIndex = 0;
			int copySize = segmentSize * byteClumps;
			if ((position + copySize) > duration)
				copySize = (duration - position) / orig.getTimeScale() * orig.getTimeScale(); // round down to nearest second
			if (copySize <= 0)
				return null;
			if (handle != null) {
				handle.unlock();
			}
			if (sd != null) {
				sd.unlock();
			}
			// yes, you may think I'm insane for shutting down QT and restarting it
			// every time we go through this loop, but trust me, there is a reason!
			// In Safari (and possibly other browsers too), the QT browser plugin
			// does some weird things when it shuts down, causing ITSM to crash the
			// next time the movieExporter is used. By reinitializing before this call,
			// we avoid the crash.
			QTSessionManager.shutdown();
			QTSessionManager.getInstance();
			try {
				allocateExporter();
			} catch (MalformedURLException e1) {
				return null;
			} catch (QTException e1) {
				return null;
			}
			try {
				handle = theMovieExp.toHandle(orig, null, position, copySize);
			} catch (QTException e1) {
				return null;
			}
			handle.lock();
			int contentBytes = (int)(copySize / (float) segmentSize * byteSize);
			offset = (int) (handle.getSize() - contentBytes);
			if (offset < -10000) {
//			 on older Safari browsers, sometimes the first call to this returns bogus data
				handle.unlock();
				try {
					handle = theMovieExp.toHandle(orig, null, position, copySize);
				} catch (QTException e) {
					return null;
				}
				handle.lock();
				contentBytes = (int)(copySize / (float) segmentSize * byteSize);
				offset = (int) (handle.getSize() - contentBytes);
			}
			int zeroFill = 0;
			if (offset < 0) {
				offset = 0;
			}
			if (bytes.length > handle.getSize()) {
				zeroFill = bytes.length - handle.getSize();
			}
			int bytesLeft = handle.getSize() - offset - byteClumpIndex * byteSize;
			if (bytesLeft < byteSize)
				return null;
			try {
				handle.copyToArray(offset + byteClumpIndex * byteSize, bytes, 0, byteSize);
				for (int i=bytes.length-zeroFill; i < bytes.length; i++) {
					bytes[i] = 0;
				}
				convertByteFormat(bytes);
			} catch (Exception e) {
				ErrorHandler.getInstance().handleError(e);
				return null;
			}
			byteClumpIndex++;
			position += (segmentSize * hopRatio);
			return bytes;
		} catch (StdQTException e) {
			return null; // don't confuse user with error messages, just bail on this song
		} catch (QTException e) {
			return null; // don't confuse user with error messages, just bail on this song
		}
	}

	public void setSegmentSizeInSeconds(double size) throws StdQTException {
		segmentSize = (int) (size * orig.getTimeScale()) + 1;
	}
	
	public int getSegmentSizeInTimeCode() {
		return segmentSize;
	}
	
	public int getSegmentPositionInTimeCode() {
		return (int) (position - (segmentSize * hopRatio));
	}
	
	public int getOrigTimeBase() {
		return origTimeBase;
	}
	
	public int getPercent() {
		return (100 * position / duration);
	}

	public void cleanup() {
		if (handle != null) {
			try {
				handle.unlock();
				sd.unlock();
			} catch (QTNullPointerException e) {
				// ignore
			}
		}
	}
	
	/**
	 * Convert byte array from twos complement to normal format
	 * @param bytes
	 */
	private void convertByteFormat(byte[] bytes) {
		int length = bytes.length;
		for (int i=0; i < length; i++) {
			bytes[i] = (byte) ((bytes[i] + 256) % 256 - 128);
		}
	}
}
