/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.mixing;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Iterator;
import java.util.Vector;

import net.jasonfreeman.itsm.analysis.Chunk;
import net.jasonfreeman.itsm.gui.ProgressManager;
import net.jasonfreeman.itsm.selection.Song;
import net.jasonfreeman.itsm.util.ErrorHandler;
import net.jasonfreeman.itsm.util.QTSessionManager;
import net.jasonfreeman.itsm.util.SettingsManager;

import quicktime.QTException;
import quicktime.QTSession;
import quicktime.io.OpenMovieFile;
import quicktime.io.QTFile;
import quicktime.std.StdQTConstants;
import quicktime.std.StdQTException;
import quicktime.std.comp.ComponentDescription;
import quicktime.std.comp.ComponentIdentifier;
import quicktime.std.movies.Movie;
import quicktime.std.movies.Track;
import quicktime.std.movies.media.AudioMediaHandler;
import quicktime.std.movies.media.Media;
import quicktime.std.movies.media.MediaHandler;
import quicktime.std.movies.media.SoundDescription;
import quicktime.std.movies.media.SoundMedia;
import quicktime.std.qtcomponents.MovieExporter;

/**
 * Main class to mix and crossfade chunks into a signature.
 * 
 * @author jason
 *
 */

public class Mixer {

    SettingsManager settings = SettingsManager.getInstance();
    
    ProgressManager progress = ProgressManager.getInstance();

    Song lastSong = null;

    Movie openMov = null;

    Track openTrack = null;

    static final int SILENCE_BUFFER_IN_TIMEBASE = 20; // in timebase
    
    int destStart = SILENCE_BUFFER_IN_TIMEBASE;
    
    int lastDur = 0;
    
    // because of QT native/Java memory management issues, we prevent
    // all of this stuff from being prematurely garbage collected by
    // keeping global variable references
    
    TweenTrack[] tweenTracks;
    
    Track[] destTracks;
    
    Track[] sourceTracks;
    
    Movie[] sourceMovs;
    
    Media[] mediaRefs;
    
    Vector tempFiles = new Vector();
    
    public Mixer() {

    }

    /**
     * Main function to mix the chunks
     * 
     * @param chunks
     * @return Movie contained mixed chunks
     */
    
    public Movie mixChunks(Chunk[] chunks) {
        try {
        		// refresh QT session to avoid possible corruption from QT browser plugin
        		// see Decoder.java for details of this hack...
        		QTSessionManager.shutdown();
        		QTSessionManager.getInstance();
            int numChunks = chunks.length;
            tweenTracks = new TweenTrack[numChunks];
            sourceTracks = new Track[numChunks];
            sourceMovs = new Movie[numChunks];
            double maxSimul = settings.getMaxSimul();
            Movie result = makeNewMovie(numChunks, maxSimul);
            double[] volumeMultipliers = setVolumeMultipliers(chunks);
            for (int i=0; i < numChunks; i++) {
                int trackNum = i + 1;
                Chunk chunk = chunks[i];
                placeChunk(chunk, trackNum, result, maxSimul, volumeMultipliers[i]);
                progress.setPercentInfo(trackNum * 100 / numChunks);
                if (!settings.isProcessing())
                		return null;
            }
            return result;
        } catch (QTException e) {
            ErrorHandler.getInstance().handleError(e);
            return null;
        }
    }

    /**
     * Initialize the signature movie
     * 
     * @param numTracks number of tracks (not including tween tracks)
     * @param maxSimul
     * @return Movie object
     * @throws QTException
     */
    private Movie makeNewMovie(int numTracks, double maxSimul) throws QTException {
        Movie result = null;
        destTracks = null;
        result = new Movie();
        destTracks = new Track[numTracks];
        for (int i = 0; i < numTracks; i++) {
            destTracks[i] = result.newTrack(0, 0, 1);
        }
        mediaRefs = new Media[numTracks];
        return result;
    }

    /**
     * Place an individual chunk in the movie and do the tween crossfade.
     * @param chunk
     * @param trackNum number of track into which it should be placed
     * @param result movie to stick it into
     * @param maxSimul
     * @param volMultiplier
     */
    private void placeChunk(Chunk chunk, int trackNum, Movie result,
            double maxSimul, double volMultiplier) {
        Song song = null;
    		try {
            // get the source track
            song = chunk.getSong();
            if (song != lastSong) {
                sourceMovs[trackNum - 1] = song.openAsMovie();
                sourceTracks[trackNum - 1] = null;
                for (int j = 1; j <= sourceMovs[trackNum - 1].getTrackCount(); j++) {
                    try {
						Track t = sourceMovs[trackNum - 1].getTrack(j);
						Media m = t.getMedia();
						MediaHandler mh = m.getHandler();
						if (mh instanceof AudioMediaHandler) {
						    sourceTracks[trackNum - 1] = t;
						    break;
						}
					} catch (QTException e) {
						// ignore, this just means that's not a track we want...
					}
                }
                if (sourceTracks[trackNum - 1] == null)
                    return;
                else {
                    openMov = sourceMovs[trackNum - 1];
                    openTrack = sourceTracks[trackNum - 1];
                    lastSong = song;
                }
            } else {
                sourceTracks[trackNum - 1] = openTrack;
                sourceMovs[trackNum - 1] = openMov;
            }

            Track destTrack = destTracks[trackNum - 1];
            destTrack.setVolume((float) (volMultiplier));
            int sourceStart = chunk.getStartTime();
            int durInSourceTime = chunk.getDur();
            int durInResultTime = durInSourceTime * result.getTimeScale() / chunk.getTimeBase() + 1;
            if ((sourceStart + durInSourceTime) >= sourceMovs[trackNum - 1].getDuration()) { // special, rare case
            		durInSourceTime = sourceMovs[trackNum - 1].getDuration() - sourceStart;
            		durInResultTime = durInSourceTime * result.getTimeScale() / chunk.getTimeBase() + 1;
            }
            if (durInResultTime == 0)
                return;
            destStart += lastDur * (1.0 / maxSimul);
            lastDur = durInResultTime;
            
            // do export / reimport hack of chunk if necessary to workaround QT6 bug
            if (buggySourceTimeBase(chunk)) {
            		File chunkFile = exportChunk(chunk, sourceMovs[trackNum - 1], sourceTracks[trackNum - 1], sourceStart, durInSourceTime);
            		tempFiles.add(chunkFile);
            		QTFile qtf = new QTFile(chunkFile);
                 OpenMovieFile omf = OpenMovieFile.asRead(qtf);
                 sourceMovs[trackNum - 1] = Movie.fromFile(omf);
                 sourceTracks[trackNum - 1] = sourceMovs[trackNum - 1].getIndTrack(1);
                 omf.close();
         		mediaRefs[trackNum - 1] = new SoundMedia(destTrack, sourceTracks[trackNum - 1]
         		                                                                .getMedia().getTimeScale());
         		sourceTracks[trackNum - 1].insertSegment(destTrack, 0, sourceMovs[trackNum - 1].getDuration(), destStart);
            } else {
            		mediaRefs[trackNum - 1] = new SoundMedia(destTrack, sourceTracks[trackNum - 1]
                    .getMedia().getTimeScale());
            		sourceTracks[trackNum - 1].insertSegment(destTrack, sourceStart, durInSourceTime, destStart);
            }
            tweenTracks[trackNum - 1] = new TweenTrack(result, destTrack, destStart, durInResultTime);
        } catch (QTException e) {
            if (song != null)
            		ErrorHandler.getInstance().handleError(e, "iTSM was unable to mix " + song.getName() + " into your signature, possibly because it was moved or deleted since the time iTSM analyzed it.");
            else
            		ErrorHandler.getInstance().handleError(e);
            try {
				result.removeTrack(destTracks[trackNum - 1]);
			} catch (StdQTException e1) {
			}
        } catch (MalformedURLException e) {
            if (song != null)
        			ErrorHandler.getInstance().handleError(e, "iTSM was unable to mix " + song.getName() + " into your signature, possibly because it was moved or deleted since the time iTSM analyzed it.");
            else
        			ErrorHandler.getInstance().handleError(e);
            try {
				result.removeTrack(destTracks[trackNum - 1]);
			} catch (StdQTException e1) {
			}
        } catch (IOException e) {
            if (song != null)
        			ErrorHandler.getInstance().handleError(e, "iTSM was unable to mix " + song.getName() + " into your signature, possibly because it was moved or deleted since the time iTSM analyzed it.");
            else
        			ErrorHandler.getInstance().handleError(e);
            try {
				result.removeTrack(destTracks[trackNum - 1]);
			} catch (StdQTException e1) {
			}
        }
    }

    /**
     * Set volumes of all the tracks
     * 
     * @param chunks
     * @return array of volume multipliers indexed the same as chunk array
     */
    private double[] setVolumeMultipliers(Chunk[] chunks) {
    		int length = chunks.length;
    		double[] volMultipliers = new double[length];
    		double biggest = 0;
    		for (int i=0; i < length; i++) {
    			volMultipliers[i] = chunks[i].getMaxAmp();
    			if (volMultipliers[i] > biggest)
    				biggest = volMultipliers[i];
    		}
    		for (int i=0; i < length; i++) {
    			// max volume is 0.1 so we can hack volume tweens up to 2550 (instead of 255) to get more resolution!
    			volMultipliers[i] = 0.06 / (volMultipliers[i] / biggest);
    		}
    		return volMultipliers;
    }
    
    /**
     * Unlock memory handles when we're done and delete any temp files.
     *
     */
    public void cleanup() {
    		if (tweenTracks != null) {
    			int length = tweenTracks.length;
    			for (int i=0; i < length; i++) {
    				TweenTrack tt = tweenTracks[i];
    				if (tt != null) {
    					tt.cleanup();
    				}
    			}
    		}
    		Iterator it = tempFiles.iterator();
    		while (it.hasNext()) {
    			File file = (File) it.next();
    			file.delete();
    		}
    }
    
    /**
     * 
     * @param chunk
     * @return true if the QT6 tween export bug applies to this file and needs the workaround
     */
    private boolean buggySourceTimeBase(Chunk chunk) {
    		return ((chunk.getTimeBase() != 600) && (QTSession.getQTMajorVersion() < 7));
    }
    
    /**
     * Export a chunk to file so we can reimport it to get around the QT6 tween export bug if needed.
     * 
     * @param in
     * @param mov
     * @param track
     * @param start
     * @param dur
     * @return
     * @throws QTException
     * @throws IOException
     */
	private File exportChunk(Chunk in, Movie mov, Track track, int start, int dur) throws QTException, IOException {
		File temp = File.createTempFile("iTSM", ".aif");
		QTFile saveFile = new QTFile(temp);
		saveFile.createMovieFile(StdQTConstants.kMoviePlayer,
				StdQTConstants.createMovieFileDeleteCurFile);
		ComponentDescription cd =
			new ComponentDescription(
				StdQTConstants.movieExportType,
				StdQTConstants.kQTFileTypeAIFF,
				StdQTConstants.soundMediaType,
				0,
				0);
		ComponentIdentifier ci = ComponentIdentifier.find(cd);
		MovieExporter exp = new MovieExporter(ci);
		SoundDescription sd = new SoundDescription(0);
		sd.setNumberOfChannels(2);
		sd.setSampleRate(44100);
		sd.setSampleSize(16);
		sd.lock();
		exp.setSampleDescription(sd, StdQTConstants.soundMediaType);
        
		exp.toFile(saveFile, mov, track, start, dur);
		sd.unlock();
		return temp;
	}

}