/*
 * iTunes Signature Maker
 * copyright (c) 2005 by Jason Freeman
 * 
 * For more information, please visit http://www.jasonfreeman.net/itsm/
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details. You should have received a copy of the GNU General
 * Public License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 */

package net.jasonfreeman.itsm.util;

import java.util.Calendar;
import java.util.Hashtable;
import java.util.TimeZone;

/**
 * Clearinghouse for all user-controllable settings. Actually, a lot of settings
 * that I chose not to put in the GUI are also here and can be controlled by changing
 * the variable initializations. Singleton class!
 * 
 * @author jason
 *
 */

public class SettingsManager extends Hashtable {

    private static SettingsManager instance = null;
    
    // selection mode constants
    public static final int TIMES_PLAYED = 0;
    public static final int LAST_DATE_PLAYED = 1;
    public static final int DATE_ADDED = 2;
    public static final int MY_RATING = 3;
    
    // limit mode constants
    public static final int LIMIT_BY_ARTIST = 0;
    public static final int LIMIT_BY_ALBUM = 1;
    public static final int LIMIT_BY_ALBUM_OR_ARTIST = 2;
    public static final int NO_LIMITS = 3;
    
    // selection variables
    private int maxSongs = 20; // number of songs to be included in signature
    
    private int selectionMode = TIMES_PLAYED;
    
    private boolean ignorePodcasts = true;
    
    private boolean ignoreVideos = true;
        
    private int limitMode = LIMIT_BY_ALBUM; // only allow 1 song in subset with matching criteria
    
    // analysis variables
    private int fftFrameSize = 16; // FFT frame size (this should probably be a constant, I think some code assumes it is always 16
    
    private int fftFrameGroupSize = 2004; // to set segment size in seconds, use commented line below instead
    
    //private int fftFrameGroupSize = (int) ((30 * Constants.ANALYSIS_SAMPLING_RATE) / fftFrameSize);
    
    private boolean fastMode = true;
    
    private double minFftMag = 0.01; // ignore frames with less energy than this
    
    private boolean varyFrameGroupSize = true; // higher-ranked songs get longer segments

    private double hopRatio = 1; // if 1, then we looked at consecutive segments, if < 1, then segments will overlap
    
    private boolean doAudioAnalysis = true;
    
    // mixing variables
    
    private double maxSimul = 6; // number of simultaneous chunks in signature
    
    private double fadePercent = 0.48; // amount of chunk which is fadein / out, e.g. 0.2 means fade in for 20% of dur and out for 20% of dur
    
    // file names and paths
    private String userHome = System.getProperty("user.home");
    
    private String fileSeparator = System.getProperty("file.separator");

    private String outDir = userHome + fileSeparator + "Desktop" + fileSeparator;
    
    private String outFile = outDir + "ITSM-" + getTimeStamp() + ".wav";

    private String xmlFile;
    
    // misc.
    
    private boolean processing = false; // should program continue processing?
    
    private String hostName = "http://works.music.columbia.edu/~jason/itsm/";
    
    // these values are changed in source distribution for security reasons
    
    private String logPath = "";
    
    private String hostAuth = "";
    
    private boolean useRemoteDb = false; // upload session, logging, and files to HTTP server

    public SettingsManager() {
        String OSname = System.getProperty("os.name");
        if (OSname.equals("Mac OS X")) {
        		xmlFile = userHome + fileSeparator + "Music" + fileSeparator + "iTunes"
        					+ fileSeparator + "iTunes Music Library.xml";
        } else
        		xmlFile = userHome + fileSeparator + "My Documents" + fileSeparator + 
        				"My Music" + fileSeparator + "iTunes" + fileSeparator + "iTunes Music Library.xml";
    }
    
    public synchronized static SettingsManager getInstance() {
        if (instance == null)
            instance = new SettingsManager();
        return instance;
    }
    
    public void setMaxSongs(int n) {
        maxSongs = n;
    }
    
    public int getMaxSongs() {
        return maxSongs;
    }
    
    public String getiTunesXmlFile() {
        return xmlFile;
    }
    
    public void setiTunesXmlFile(String xmlFile) {
        this.xmlFile = xmlFile;
    }
    
    public String getOutFile() {
        return outFile;
    }
    
    public String getOutDir() {
    		return outDir;
    }
    
    public void setOutFile(String outFile) {
    		this.outFile = outFile;
    }
    
    public int getFftFrameSize() {
        return fftFrameSize;
    }
    
    /**
     * 
     * @param rank 0 is first-place ranking, 1 is last-place ranking
     * @return weighted FFtFrameGroupSize if segment lengths vary, o/w actual size
     */
    public int getFftFrameGroupSize(float rank) {
    		if (!varyFrameGroupSize)
    			return fftFrameGroupSize;
    		return (int) (fftFrameGroupSize * 0.5 + (1.0 - rank) * fftFrameGroupSize);
    }
    
    public double getMaxSimul() {
        return maxSimul;
    }
    
    public void setMaxSimul(double maxSimul) {
    		this.maxSimul = maxSimul;
    }
    
    public double getFadePercent() {
        return fadePercent;
    }
    
    public int getSelectionMode() {
        return selectionMode;
    }
    
    public void setSelectionMode(int selectionMode) {
        this.selectionMode = selectionMode;
    }
    
    public double getWindowSizeInSeconds(float rank) {
    		return((fftFrameSize * getFftFrameGroupSize(rank)) / (double) Constants.ANALYSIS_SAMPLING_RATE);
    }

    public void setWindowSizeInSeconds(double seconds) {
    		fftFrameGroupSize = (int) ((seconds * Constants.ANALYSIS_SAMPLING_RATE) / fftFrameSize);
    }
    
    public void setProcessing(boolean running) {
    		this.processing = running;
    }
    
    public boolean isProcessing() {
    		return processing;
    }
    
    private String getTimeStamp() {
        Calendar cal = Calendar.getInstance(TimeZone.getDefault());
        String DATE_FORMAT = "yyyy-MM-dd-HH-mm-ss";
        java.text.SimpleDateFormat sdf = 
              new java.text.SimpleDateFormat(DATE_FORMAT);
        sdf.setTimeZone(TimeZone.getDefault());          
        return(sdf.format(cal.getTime()));
    }

	public boolean isFastMode() {
		return fastMode;
	}

	public void setFastMode(boolean fastMode) {
		this.fastMode = fastMode;
	}
	
	public double getMinFftMag() {
		return minFftMag;
	}
	
	public static void shutdown() {
		instance = null;
	}
	
	public String updateOutFileName() {
		outFile = userHome + "/Desktop/ITSM-" + getTimeStamp() + ".aiff";
		return outFile;
	}
	
	public String getHostName() {
		return hostName;
	}
	
	public String getHostAuth() {
		return hostAuth;
	}
	
	public String getLogPath() {
		return logPath;
	}
	
	public boolean useRemoteDb() {
		return useRemoteDb;
	}

	public boolean isIgnorePodcasts() {
		return ignorePodcasts;
	}

	public void setIgnorePodcasts(boolean ignorePodcasts) {
		this.ignorePodcasts = ignorePodcasts;
	}

	public boolean isIgnoreVideos() {
		return ignoreVideos;
	}

	public void setIgnoreVideos(boolean ignoreVideos) {
		this.ignoreVideos = ignoreVideos;
	}
	
	public double getHopRatio() {
		return hopRatio;
	}

	public int getLimitMode() {
		return limitMode;
	}

	public void setLimitMode(int limitMode) {
		this.limitMode = limitMode;
	}

	public boolean isDoAudioAnalysis() {
		return doAudioAnalysis;
	}

	public void setDoAudioAnalysis(boolean doAudioAnalysis) {
		this.doAudioAnalysis = doAudioAnalysis;
	}

}
